//////////////////////////////////////////////////////////////////////
//                                                                  //
// console.pas: Console and scripting system                        //
//   Manages the console, log, and configuration managment          //
//                                                                  //
// The contents of this file are subject to the Bottled Light       //
// Public License Version 1.0 (the "License"); you may not use this //
// file except in compliance with the License. You may obtain a     //
// copy of the License at http://www.bottledlight.com/BLPL/         //
//                                                                  //
// Software distributed under the License is distributed on an      //
// "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   //
// implied. See the License for the specific language governing     //
// rights and limitations under the License.                        //
//                                                                  //
// The Original Code is the Mappy VM User Interface, released       //
// April 1st, 2003. The Initial Developer of the Original Code is   //
// Bottled Light, Inc. Portions created by Bottled Light, Inc. are  //
// Copyright (C) 2001-2003 Bottled Light, Inc. All Rights Reserved. //
//                                                                  //
// Author(s):                                                       //
//   Michael Noland (joat), michael@bottledlight.com                //
//                                                                  //
// Changelog:                                                       //
//   1.0: First public release (April 1st, 2003)                    //
//                                                                  //
// Notes:                                                           //
//   FontToString and BooleanToString should probably be moved      //
//   somewhere else.                                                //
//                                                                  //
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
unit console; ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
interface ////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

uses
  Classes, Graphics, SysUtils, Math;

//////////////////////////////////////////////////////////////////////

type
  TConsoleCallback = function (params: string): string;
  TLogCallback = procedure (newLines: integer) of object;

//////////////////////////////////////////////////////////////////////

// logWrite prints a string to the log/console, you must provide
// line returns yourself
procedure logWrite(st: string);

// logWriteLn prints a string to the console and appends a line return
procedure logWriteLn(st: string);

// logAddCommand adds a function to the console namespace.  A console
// command or callback takes its parameters in the form of a string, and
// returns status information or an error code if desired.  For any static
// valued function (one prefixed with a $, to be saved on program exit),
// an input parameter of '' should return a string that can be passed in
// at a later date as a parameter to restore the state, i.e. the following
// should be an identity function:
//    logSetValue(funcName, logGetValue(funcName))
// As logGetValue with a parameter of type function returns the value that
// the function returns when passed ''
procedure logAddCommand(name: string; callback: TConsoleCallback); overload;
procedure logAddCommand(name: string; callback: TConsoleCallback; help: string); overload;
procedure logAddCommand(name: string; callback: TConsoleCallback; value, help: string); overload;

// Adds an integer|string|boolean
procedure logAddInteger(name: string; var d: integer); overload;
procedure logAddInteger(name: string; var d: integer; help: string); overload;
procedure logAddInteger(name: string; var d: integer; value, help: string); overload;
procedure logAddString(name: string; var d: string); overload;
procedure logAddString(name: string; var d: string; help: string); overload;
procedure logAddString(name: string; var d: string; value, help: string); overload;
procedure logAddBoolean(name: string; var d: boolean); overload;
procedure logAddBoolean(name: string; var d: boolean; help: string); overload;
procedure logAddBoolean(name: string; var d: boolean; value, help: string); overload;

// Get the value of a console variable, or the result of a console function
// called with no parameters
function logGetValue(name: string): string; overload;
function logGetValue(index: integer): string; overload;

// Set a console varable's value, or call a console function with a parameter
// list of value
procedure logSetValue(name, value: string); overload;
procedure logSetValue(index: integer; value: string); overload;

// Process a command as if it were typed into the console
procedure logProcessCommand(st: string);
procedure logProcessCommandSilent(st: string);

// Set the callback function for a log observer
procedure logSetCallback(callback: TLogCallback);

// Retrieve a range of entries from the log
procedure logGetEntries(last, count: integer; list: TStrings);

// Returns the number of entries in the log
function logNumEntries: integer;

// Returns true if there are new entries in the log since the
// last time logGetEntries was called, otherwise false.
function logBeenUpdated: boolean;

// Loads and executes a sequence of commands on lines in the passed
// in file or string list
procedure logExecuteScript(filename: string); overload;
procedure logExecuteScript(script: TStringList); overload;

// Saves any console variables whose names are prefixed with a '$',
// and the null result (result returned when a function is called
// with no parameters) for any console functions prexied with '$'
// to a file in a format suitable for loading via logExecuteScript
procedure logSaveSettings(filename: string); overload;
procedure logSaveSettings(script: TStringList); overload;

procedure logSaveLog(filename: string);
procedure logClear;

//////////////////////////////////////////////////////////////////////

procedure StringToFont(sFont: string; Font: TFont);
function FontToString(Font: TFont): string;

function BooleanToString(value: boolean): string;
function StringToBoolean(value: string): boolean;

function CutLeft(var st: string): string;
function ThwackString(ch: char; var st: string): string;
function CutAnyLeft(var st: string; ch: char): string;

//////////////////////////////////////////////////////////////////////
implementation ///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

//uses
//  dbgExpressions;

//////////////////////////////////////////////////////////////////////

type
  TConsoleType = (ctFunc, ctInteger, ctString, ctBoolean);

  TConsoleEntry = class
    myType: TConsoleType;
    myData: pointer;
    help: string;
    constructor Create(typ: TConsoleType; data: pointer); overload;
    constructor Create(typ: TConsoleType; data: pointer; helpInfo: string); overload;
    procedure SetValue(st: string);
    function GetHelp: string;
    function GetValue: string;
  end;

//////////////////////////////////////////////////////////////////////

const
  cTypeNames: array[TConsoleType] of string =
    ('function', 'integer', 'string', 'boolean');

//////////////////////////////////////////////////////////////////////

var
  currentString: string;
  log: TStringList;
  maxLogSize: integer;
  cvars: TStringList;
  logViewer: TLogCallback;
  beenUpdated: boolean;

//////////////////////////////////////////////////////////////////////

const
  csfsBold = '|Bold';
  csfsItalic = '|Italic';
  csfsUnderline = '|Underline';
  csfsStrikeout = '|Strikeout';

//////////////////////////////////////////////////////////////////////

function CutLeft(var st: string): string;
var
  i: integer;
begin
  i := Pos(' ', st);
  if i > 0 then begin
    Result := Copy(st, 1, i-1);
    Delete(st, 1, i);
  end else begin
    Result := st;
    st := '';
  end;
end;

//////////////////////////////////////////////////////////////////////

function ThwackString(ch: char; var st: string): string;
var
  i: integer;
begin
  i := Pos(ch, st);
  if i > 0 then begin
    Result := Copy(st, i+1, Length(st)-i);
    Delete(st, i, Length(st)-i+1);
  end else
    Result := '';
end;

//////////////////////////////////////////////////////////////////////

function CutAnyLeft(var st: string; ch: char): string;
var
  i: integer;
begin
  i := Pos(ch, st);
  if i > 0 then begin
    Result := Copy(st, 1, i-1);
    Delete(st, 1, i);
  end else begin
    Result := st;
    st := '';
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure StringToFont(sFont: string; Font: TFont);
var
  p: integer;
  sStyle: string;
begin
  with Font do begin
    // get font name
    p := Pos( ',', sFont );
    Name := Copy(sFont, 2, p-3);
    Delete(sFont, 1, p);

    // get font size
    p := Pos(',', sFont);
    Size := StrToInt(Copy(sFont, 2, p-2));
    Delete(sFont, 1, p);

    // get font style
    p := Pos( ',', sFont);
    sStyle := '|' + Copy(sFont, 3, p-4);
    Delete(sFont, 1, p);

    // get font color
    Color := StringToColor(Copy(sFont, 3, Length( sFont ) - 3));

    Style := [];
    if Pos(csfsBold, sStyle) > 0 then Style := Style + [fsBold];
    if Pos(csfsItalic, sStyle) > 0 then Style := Style + [fsItalic];
    if Pos(csfsUnderline, sStyle) > 0 then Style := Style + [fsUnderline];
    if Pos(csfsStrikeout, sStyle) > 0 then Style := Style + [fsStrikeout];
  end;
end;

//////////////////////////////////////////////////////////////////////

function FontToString(Font: TFont): string;
var
  sStyle: string;
begin
  with Font do begin
    sStyle := '';

    if fsBold in Style then sStyle := sStyle + csfsBold;
    if fsItalic in Style then sStyle := sStyle + csfsItalic;
    if fsUnderline in Style then sStyle := sStyle + csfsUnderline;
    if fsStrikeout in Style then sStyle := sStyle + csfsStrikeout;

    if (Length(sStyle) > 0) and (sStyle[1] = '|') then
      sStyle := Copy(sStyle, 2, Length(sStyle) - 1);

    Result := Format('"%s", %d, [%s], [%s]', [Name, Size, sStyle, ColorToString(Color)]);
  end;
end;

//////////////////////////////////////////////////////////////////////

function BooleanToString(value: boolean): string;
begin
  if value then Result := 'true' else Result := 'false';
end;

//////////////////////////////////////////////////////////////////////

function StringToBoolean(value: string): boolean;
begin
  value := Uppercase(value);
  if value = 'TRUE' then
    Result := true
  else if value = 'FALSE' then
    Result := false
  else
    Result := StrToIntDef(value, 0) <> 0;
end;

//////////////////////////////////////////////////////////////////////

function CFuncSetValue(params: string): string;
begin
  // Just an alias for a normal set command only its silent
  logProcessCommandSilent(params);
  Result := '';
end;

//////////////////////////////////////////////////////////////////////

function CFuncList(params: string): string;
var
  i: integer;
begin
  // List all the entries in the namespace
  for i := 0 to cvars.Count - 1 do
    logWriteLn('  ' + cvars.strings[i] + ': ' + cTypeNames[TConsoleEntry(cvars.objects[i]).myType]);
  Result := '';
end;

//////////////////////////////////////////////////////////////////////

function CFuncHelp(params: string): string;
var
  st: string;
  index: integer;
begin
  // Load the help
  if params = '' then params := 'help';
  index := cvars.IndexOf(lowercase(params));
  if index > -1 then
    st := TConsoleEntry(cvars.Objects[index]).help
  else
    st := '';

  // Display the help
  if st <> '' then
    logWriteLn(params + ': ' + st)
  else
    logWriteLn('No help entry found for ' + params);

  // Return the help, for whatever reason
  Result := '';
end;

//////////////////////////////////////////////////////////////////////

function CFuncClear(params: string): string;
begin
  logClear;
  if Assigned(logViewer) then logViewer(0);
end;

//////////////////////////////////////////////////////////////////////

function FirstToken(st: string): string;
var
  i: integer;
begin
  i := Pos(' ', st);
  if i = 0 then i := Length(st)+1;
  Result := Copy(st, 1, i-1);
  beenUpdated := true;
end;

//////////////////////////////////////////////////////////////////////

procedure logWriteLn(st: string);
begin
  log.Add(currentString + st);
{  while log.Count > maxLogSize do begin
    log.Delete(0);
    log.Delete(0);
    log.Delete(0);
    log.Delete(0);
    log.Delete(0);
    log.Delete(0);
  end;}
  currentString := '';
  if Assigned(logViewer) then logViewer(1);
  beenUpdated := true;
end;

//////////////////////////////////////////////////////////////////////

procedure logWrite(st: string);
var
  p: integer;
begin
  p := Pos(#10, st);
  while p > 0 do begin
    logWriteLn(Copy(st, 1, p-1));
    Delete(st, 1, p);
    p := Pos(#10, st);
  end;
  currentString := currentString + st;
  if Assigned(logViewer) then logViewer(0);
end;

//////////////////////////////////////////////////////////////////////

procedure logAddCommand(name: string; callback: TConsoleCallback);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctFunc, @callback));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddCommand(name: string; callback: TConsoleCallback; help: string);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctFunc, @callback, help));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddCommand(name: string; callback: TConsoleCallback; value, help: string);
begin
  logSetValue(cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctFunc, @callback, help)), value);
end;

//////////////////////////////////////////////////////////////////////

procedure logAddInteger(name: string; var d: integer);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctInteger, @d));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddInteger(name: string; var d: integer; help: string);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctInteger, @d, help));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddInteger(name: string; var d: integer; value, help: string);
begin
  logSetValue(cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctInteger, @d, help)), value);
end;

//////////////////////////////////////////////////////////////////////

procedure logAddString(name: string; var d: string);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctString, @d));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddString(name: string; var d: string; help: string);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctString, @d, help));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddString(name: string; var d: string; value, help: string);
begin
  logSetValue(cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctString, @d, help)), value);
end;

//////////////////////////////////////////////////////////////////////

procedure logAddBoolean(name: string; var d: boolean);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctBoolean, @d));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddBoolean(name: string; var d: boolean; help: string);
begin
  cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctBoolean, @d, help));
end;

//////////////////////////////////////////////////////////////////////

procedure logAddBoolean(name: string; var d: boolean; value, help: string);
begin
  logSetValue(cvars.AddObject(Lowercase(name), TConsoleEntry.Create(ctBoolean, @d, help)), value);
end;

//////////////////////////////////////////////////////////////////////

procedure logProcessCommand(st: string);
var
  cmd, s2: string;
  i, j: integer;
begin
  // Take the first word, lowercase it and try to match it to things
  st := TrimLeft(st);
  cmd := Lowercase(FirstToken(st));
  if cmd = '' then Exit;

  case cmd[1] of
    '=': begin
      // Do an expression evaluation
      i := Pos('=', st);
      st := Copy(st, i+1, Length(st)-i);
//            logWriteLn(IntToStr(EvaluateExpression(st)));
    end;
    ':': begin
      // Enter a program statement
    end;
  else
    // Do a normal command or lookup
    if cmd[1] = '/' then Delete(cmd, 1, 1);
    i := cvars.IndexOf(cmd);
    if i > -1 then begin
      j := Pos(' ', st);

      // If its a variable and we have something to assign to it, do so
      if j > 0 then
        logSetValue(i, Copy(st, j+1, Length(st)-j));

      // Get the value / execute the function
      if (TConsoleEntry(cvars.Objects[i]).myType <> ctFunc) or (j = 0) then
        s2 := logGetValue(i)
      else
        s2 := '';

      // Unless its a function that returned '' then print the value/result
      if (TConsoleEntry(cvars.Objects[i]).myType <> ctFunc) or (s2 <> '') then
        logWriteLn(cmd + ' is ' + s2);
    end else
      logWriteLn('"' + st + '"');
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure logProcessCommandSilent(st: string);
var
  cmd: string;
  i, j: integer;
begin
  // Take the first word, lowercase it and try to match it to things
  st := TrimLeft(st);
  cmd := Lowercase(FirstToken(st));
  if cmd = '' then Exit;
  if cmd[1] = '/' then Delete(cmd, 1, 1);
  i := cvars.IndexOf(cmd);
  if i > -1 then begin
    j := Pos(' ', st);
    if j > 0 then logSetValue(i, Copy(st, j+1, Length(st)-j));
    if (TConsoleEntry(cvars.Objects[i]).myType <> ctFunc) or (j = 0) then logGetValue(i)
  end;
end;

//////////////////////////////////////////////////////////////////////

function logGetValue(name: string): string;
var
  index: integer;
begin
  index := cvars.IndexOf(name);
  if index > -1 then
    Result := TConsoleEntry(cvars.Objects[index]).GetValue
  else
    Result := '';
end;

//////////////////////////////////////////////////////////////////////

function logGetValue(index: integer): string;
begin
  Result := TConsoleEntry(cvars.Objects[index]).GetValue;
end;

//////////////////////////////////////////////////////////////////////

procedure logSetValue(name, value: string);
var
  index: integer;
begin
  index := cvars.IndexOf(name);
  if index > -1 then
    TConsoleEntry(cvars.Objects[index]).SetValue(value);
end;

//////////////////////////////////////////////////////////////////////

procedure logSetValue(index: integer; value: string);
begin
  TConsoleEntry(cvars.Objects[index]).SetValue(value);
end;

//////////////////////////////////////////////////////////////////////
// TConsoleEntry /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

constructor TConsoleEntry.Create(typ: TConsoleType; data: pointer);
begin
  myType := typ;
  myData := data;
  help := '';
end;

//////////////////////////////////////////////////////////////////////

constructor TConsoleEntry.Create(typ: TConsoleType; data: pointer; helpInfo: string);
begin
  myType := typ;
  myData := data;
  help := helpInfo;
end;

//////////////////////////////////////////////////////////////////////

function TConsoleEntry.GetHelp: string;
begin
  Result := help;
end;

//////////////////////////////////////////////////////////////////////

function TConsoleEntry.GetValue: string;
const
  bools: array[false..true] of string = ('false', 'true');
begin
  case myType of
    ctFunc: Result := TConsoleCallback(myData)('');
    ctInteger: Result := IntToStr(integer(myData^));
    ctString: Result := string(myData^);
    ctBoolean: Result := bools[boolean(myData^)];
  else
    Result := 'error: invalid type';
  end;
end;

//////////////////////////////////////////////////////////////////////

function ExpandLine(st: string): string;
var
  i, j: integer;
  left, right: string;
begin
  i := Pos('%', st);
  while i > 0 do begin
    // l33t string magik :P
    left := Copy(st, 1, i-1);
    Delete(st, 1, i);
    j := Pos('%', st);
    if j = 0 then j := Length(st)+1;
    right := Copy(st, 1, j-1);
    Delete(st, 1, j);

    // Substitution
    i := Pos('$', st);
    st := left + logGetValue(Lowercase(right)) + right;
  end;
  Result := st;
end;

//////////////////////////////////////////////////////////////////////

procedure TConsoleEntry.SetValue(st: string);
begin
  case myType of
    ctFunc: begin
      // Call a function
      TConsoleCallback(myData)(st);
    end;
    ctInteger: begin
      // Process a command line to set the value of an integer variable
      st := Uppercase(ExpandLine(FirstToken(st)));
      if (st = 'TRUE') or (st = 'ON') or (st = 'YES') then
        integer(myData^) := 1
      else if (st = 'FALSE') or (st = 'OFF') or (st = 'NO') then
        integer(myData^) := 0
      else
        integer(myData^) := StrToIntDef(st, integer(myData^));
    end;
    ctString: begin
      // Process the command line to set a string variable
      string(myData^) := st;
    end;
    ctBoolean: begin
      // Process the command line to set a boolean variable
      st := Uppercase(ExpandLine(FirstToken(st)));
      if (st = 'TRUE') or (st = 'ON') or (st = 'YES') then
        boolean(myData^) := true
      else if (st = 'FALSE') or (st = 'OFF') or (st = 'NO') then
        boolean(myData^) := false
      else
        boolean(myData^) := StrToIntDef(st, Ord(boolean(myData^))) <> 0;
    end;
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure logSetCallback(callback: TLogCallback);
begin
  logViewer := callback;
end;

//////////////////////////////////////////////////////////////////////

procedure logGetEntries(last, count: integer; list: TStrings);
var
  i: integer;
begin
  if count = 0 then Exit;
  for i := Max(last-count, 1) to Min(last, log.Count) do
    list.Add(log.Strings[i-1]);
  if currentString <> '' then list.Add(currentString);
  beenUpdated := false;
end;

//////////////////////////////////////////////////////////////////////

function logNumEntries: integer;
begin
  Result := log.Count;
  if currentString <> '' then Inc(Result);
end;

//////////////////////////////////////////////////////////////////////

function logBeenUpdated: boolean;
begin
  Result := beenUpdated;
end;

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

procedure InitConsole;
begin
  if not Assigned(log) then begin
    // Create the log file and associated stuff
    log := TStringList.Create;
    log.Add('Mappy VM: Logging started on ' + DateTimeToStr(Now));
    maxLogSize := 1 shl 18;
    log.Capacity := maxLogSize shr 1;
    currentString := '';

    // Create the console and associated stuff
    cvars := TStringList.Create;
    logViewer := nil;
    beenUpdated := false;
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure DestroyConsole;
var
  i: integer;
begin
  log.Free;
  for i := 0 to cvars.Count-1 do
    TConsoleEntry(cvars.Objects[i]).Free;
  cvars.Free;
end;

//////////////////////////////////////////////////////////////////////

procedure logExecuteScript(filename: string);
var
  list: TStringList;
begin
  if FileExists(filename) then begin
    // Load the file into a string list and use the other function
    list := TStringList.Create;
    list.LoadFromFile(filename);
    logExecuteScript(list);
    list.Free;
  end else
    logWriteLn('ERROR: Script ' + filename + ' could not be found');
end;

//////////////////////////////////////////////////////////////////////

procedure logExecuteScript(script: TStringList);
var
  index: integer;
begin
  // Load the script and execute it
  for index := 0 to script.Count-1 do
    logProcessCommand(script.Strings[index]);
end;

//////////////////////////////////////////////////////////////////////

procedure logSaveSettings(filename: string);
var
  list: TStringList;
begin
  // Create a string list, fill it, and save it
  list := TStringList.Create;
  logSaveSettings(list);
  list.SaveToFile(filename);
  list.Free;
end;

//////////////////////////////////////////////////////////////////////

procedure logSaveSettings(script: TStringList);
var
  index: integer;
  name: string;
begin
  // Save any variables with a '$' prefix to a config file
  for index := 0 to cvars.Count-1 do begin
    name := cvars.Strings[index];
    if (name <> '') and (name[1] = '$') then begin
      if TConsoleEntry(cvars.Objects[index]).myType = ctFunc then
        script.Add(name + ' ' + logGetValue(index))
      else
        script.Add('/set ' + name + ' ' + logGetValue(index));
    end;
  end;
end;

//////////////////////////////////////////////////////////////////////

procedure logSaveLog(filename: string);
begin
  log.Add(currentString);
  log.SaveToFile(filename);
  if log.Count > 0 then log.Delete(log.Count-1);
end;

//////////////////////////////////////////////////////////////////////

procedure logClear;
begin
  log.Clear;
end;

//////////////////////////////////////////////////////////////////////

initialization
  InitConsole;
  logAddCommand('set', CFuncSetValue, 'Sets a console object to a specified value');
  logAddCommand('list', CFuncList, 'Lists all console objects in the current namespace');
  logAddCommand('help', CFuncHelp, 'Displays help for a console object if available');
  logAddCommand('clear', CFuncClear, 'Clears the console');
finalization
//  logSaveSettings(ExtractFilePath(ParamStr(0)) + 'setup.cfg');
  DestroyConsole;
end.

//////////////////////////////////////////////////////////////////////
